/*
 **  EPIwm
 **  Copyright (C) 1999 Julien Mulot, Thomas Poindessous, Stephane Rault
 **
 **  This program is free software; you can redistribute it and/or modify
 **  it under the terms of the GNU General Public License as published by
 **  the Free Software Foundation; either version 2 of the License, or
 **  (at your option) any later version.
 **
 **  This program is distributed in the hope that it will be useful,
 **  but WITHOUT ANY WARRANTY; without even the implied warranty of
 **  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 **  GNU General Public License for more details.
 **
 **  You should have received a copy of the GNU General Public License
 **  along with this program; if not, write to the Free Software
 **  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 **
 ** window.c
 ** 
 ** Made by (Julien Mulot)
 ** Login   <mulot@epita.fr>
 ** 
 ** Started on  Fri Jan  8 21:13:01 1999 Julien Mulot
 ** Last update Tue Sep 21 11:24:21 1999 mulot
 */

#include "epiwm.h"
#include <X11/Xutil.h>
#include <X11/extensions/shape.h>

int		test_shape(win)
Window		win;
{
  int		xws;
  int		yws;
  int		xbs;
  int		ybs;
  unsigned	wws;
  unsigned	hws;
  unsigned	wbs;
  unsigned	hbs;
  int		boundingShaped;
  int		clipShaped;

  XShapeSelectInput(gl_dpy, win, ShapeNotifyMask);
  XShapeQueryExtents(gl_dpy, win, &boundingShaped, &xws, &yws, &wws,
		     &hws, &clipShaped, &xbs, &ybs, &wbs, &hbs);
  return (boundingShaped);
}

void			add_shape(wg)
t_win			*wg;
{
  XRectangle		rect[4];
  XWindowAttributes	attrib;
  unsigned short	nb_rect;
  
  /* nb_rect nb de cotes */
  nb_rect = 4;
  XGetWindowAttributes(gl_dpy, wg->parent, &attrib);
  rect[0].x = 0;
  rect[0].y = 0;
  rect[0].width = attrib.width ;
  rect[0].height = wg->border_size + wg->titlebar_height +
    wg->titlebar_border*2;
  rect[1].x = 0;
  rect[1].y = 0;
  rect[1].width = wg->border_size + wg->bf_border;
  rect[1].height = attrib.height;
  rect[2].x = attrib.width - wg->border_size - wg->bf_border;
  rect[2].y = 0;
  rect[2].width = wg->border_size + wg->bf_border;
  rect[2].height = attrib.height;
  rect[3].x = 0;
  rect[3].y = attrib.height - wg->border_size - wg->bf_border;
  rect[3].width = attrib.width;
  rect[3].height = wg->border_size + wg->bf_border;   
  XGetWindowAttributes(gl_dpy, wg->win, &attrib);
  XShapeCombineShape(gl_dpy, wg->parent, ShapeBounding,
		     wg->border_size + attrib.border_width
		     + wg->bf_border,
		     wg->border_size + attrib.border_width
		     + wg->titlebar_height
		     + 2*wg->titlebar_border,
		     wg->win, ShapeBounding, ShapeSet);
  if (wg->border_size != 0 && wg->titlebar_height != 0)
    XShapeCombineRectangles(gl_dpy, wg->parent, ShapeBounding,
			    0, 0, rect, nb_rect,
			    ShapeUnion, Unsorted);     
}

t_win			*add_border(win)
Window			win;
{     
  t_win			*new_win;
  XWindowAttributes	attrib;
  XSetWindowAttributes	attr;  
  char			*name;
  XWMHints		*wmhints;
  Window		transient;
  
  XGetWindowAttributes(gl_dpy, win, &attrib);
  if (!attrib.override_redirect)
    {
      new_win = (t_win *) malloc(sizeof(t_win));
      new_win->win = win;
      XSetWindowBorderWidth(gl_dpy, win, 0);
      XGetWindowAttributes(gl_dpy, win, &attrib);
      XAddToSaveSet(gl_dpy, win);
      XGrabServer(gl_dpy);
      XUnmapWindow(gl_dpy, win);
      XFetchName(gl_dpy, win, &name);
      check_app_prop(new_win, name);
      motif_check(new_win);
      check_properties(new_win);
      new_win->is_win_shape = test_shape(new_win->win);
      new_win->parent = add_parent(new_win, &attrib);
      new_win->border_frame = add_border_frame(new_win, new_win->parent);
      new_win->titlebar_win = add_titlebar_win(new_win, new_win->parent);
      XStoreName(gl_dpy, new_win->parent, name);
      if ((new_win->prop & NO_TITLE) == 0)
	XStoreName(gl_dpy, new_win->titlebar_win, name);
      free(name);
      XGetIconName(gl_dpy, win, &name);
      XSetIconName(gl_dpy, new_win->parent, name);
      free(name);
      if (gl_epiwin.align_btn_max)
	new_win->button_max = build_button(new_win,
					   gl_epiwin.align_btn_max,
					   gl_epiwin.pos_btn_max);
      if (gl_epiwin.align_btn_max_height)
	new_win->button_max_height = 
	  build_button(new_win, 
		       gl_epiwin.align_btn_max_height,
		       gl_epiwin.pos_btn_max_height);
      if (gl_epiwin.align_btn_max_width)
	new_win->button_max_width = build_button(new_win,
						 gl_epiwin.align_btn_max_width,
						 gl_epiwin.pos_btn_max_width);
      
      if (gl_epiwin.align_btn_iconify)
	new_win->button_iconify = build_button(new_win,
					       gl_epiwin.align_btn_iconify,
					       gl_epiwin.pos_btn_iconify);
      if (gl_epiwin.align_btn_shade)
	new_win->button_shade = build_button(new_win,
					     gl_epiwin.align_btn_shade,
					     gl_epiwin.pos_btn_shade);
      if (gl_epiwin.align_btn_close)
	new_win->button_close = build_button(new_win,
					     gl_epiwin.align_btn_close,
					     gl_epiwin.pos_btn_close);
      if (gl_epiwin.align_btn_destroy)
	new_win->button_destroy = build_button(new_win,
					       gl_epiwin.align_btn_destroy,
					       gl_epiwin.pos_btn_destroy);
      add_elt(new_win);
      attr.event_mask = PropertyChangeMask | ResizeRedirectMask
	| StructureNotifyMask | KeyPressMask | PointerMotionMask;
      attr.do_not_propagate_mask = ButtonPressMask | ButtonReleaseMask;
      XChangeWindowAttributes(gl_dpy, new_win->win,
			      CWEventMask | CWDontPropagate, &attr);
      XReparentWindow(gl_dpy, win, new_win->border_frame,
		      new_win->border_size, new_win->border_size);
      grab_all_shortcuts(win);
      wmhints = XGetWMHints(gl_dpy, win);
      window_state(new_win, "inactive");
      if (wmhints)
	{
	  if ((wmhints->flags & StateHint) &&
	      (wmhints->initial_state == IconicState))
	    win_iconify(new_win, 1, 0);
	  else if (attrib.map_state == IsViewable)
	    {
	      if ((gl_epiwin.smart_placement) && (!gl_epiwin.first_start))
		placement_smart(new_win);
	      window_group_map(new_win);
	      window_state(new_win, "inactive");
	      catch_win_mouse();
	    }
	}
      else if (attrib.map_state == IsViewable)
	{
	  if ((gl_epiwin.smart_placement) && (!gl_epiwin.first_start))
	    placement_smart(new_win);
	  window_group_map(new_win);
	  window_state(new_win, "inactive");
	  catch_win_mouse();
	}
      if (XGetTransientForHint(gl_dpy, new_win->win, &transient))
	{
	  if (transient != gl_scr_info.root_win)
	    map_request(transient);
	}
      XUngrabServer(gl_dpy);
      XSync(gl_dpy, 1);
      return(new_win);
    }
  else
    return(NULL);
}

void			handle_window(win)
Window			win;
{
  XWindowAttributes	attrib;
  
  XGetWindowAttributes(gl_dpy, win, &attrib);
  if (attrib.map_state == IsViewable)
    add_border(win);
}

t_win		*create_win(win)
Window		win;
{
  t_win		*tmp;
  
  tmp = add_border(win);
  return (tmp);
}

Window			add_parent(wg, attrib)
t_win			*wg;
XWindowAttributes	*attrib;
{
  Window		temp;
  XSetWindowAttributes	attr;
  int			width;
  int			height;
  
  width = 0;
  height = 0;
  if (wg->prop & NO_TITLE)
    {
      width = attrib->width + (wg->border_size + wg->bf_border) * 2;
      height = attrib->height + (wg->border_size + wg->bf_border) * 2;
    }
  else if (gl_epiwin.titlebar.position == TOP
	   || gl_epiwin.titlebar.position == BOTTOM)
    {
      width = attrib->width + (wg->border_size + wg->bf_border) * 2;
      height = attrib->height +	(wg->border_size + wg->titlebar_border)*2
	+ wg->titlebar_height + wg->bf_border;
    }
  else if (gl_epiwin.titlebar.position == LEFT
	   || gl_epiwin.titlebar.position == RIGHT)
    {
      width = attrib->width + wg->titlebar_height + wg->titlebar_border * 2
	+ wg->border_size * 2 + wg->bf_border;
      height = attrib->height +	(wg->border_size + wg->bf_border)*2;
    }
  temp = XCreateSimpleWindow(gl_dpy,
			     gl_scr_info.root_win,
			     attrib->x, attrib->y, width, height, 0,
			     BlackPixel(gl_dpy, gl_scr_info.screen),
			     WhitePixel(gl_dpy, gl_scr_info.screen));
  attr.event_mask = /*ExposureMask | */EnterWindowMask | LeaveWindowMask
    | ButtonPressMask | ButtonReleaseMask | SubstructureNotifyMask
      | PropertyChangeMask | SubstructureRedirectMask | KeyPressMask;
  XChangeWindowAttributes(gl_dpy, temp , CWEventMask, &attr);
  return (temp);
}

Window			build_button(win, align, pos)
t_win			*win;
char			align;
int			pos;
{
  t_window_btn		tmp;
  XSetWindowAttributes	attrib;  
  Window		w_tmp;
  
  if (win->prop & NO_TITLE)
    return (0);
  if ((gl_epiwin.titlebar.position == TOP) 
      || (gl_epiwin.titlebar.position == BOTTOM))
    {
      tmp.x = set_btn_position(win, align, pos);
      tmp.y = (win->titlebar_height - gl_epiwin.btn_height) / 2;
    }
  else
    {
      tmp.x = (win->titlebar_height - gl_epiwin.btn_width) / 2;
      tmp.y = set_btn_position(win, align, pos);
    }
  tmp.width = gl_epiwin.btn_width;
  tmp.height = gl_epiwin.btn_height ;
  tmp.bgcolor = BlackPixel(gl_dpy, gl_scr_info.screen);
  tmp.fgcolor = gl_epiwin.titlebar.color_active.pixel;
  tmp.parent = win->titlebar_win;
  tmp.cursor = gl_epiwin.normal_cursor;
  tmp.border = 0;
  w_tmp = add_btn(&tmp);
  attrib.event_mask = ButtonPressMask | ButtonReleaseMask
    /* | ExposureMask */;
  XChangeWindowAttributes(gl_dpy, w_tmp, CWEventMask, &attrib);
  return (w_tmp);
}

Window			add_border_frame(wg, parent)
t_win			*wg;
Window			parent;
{
  t_window_btn		tmp;
  XWindowAttributes	attrib;
  XSetWindowAttributes	attr;
  Window		w_tmp;
  int			width;
  int			height;
  int			x;
  int			y;
  
  x = 0;
  y = 0;
  width = 0;
  height = 0;
  XGetWindowAttributes(gl_dpy, parent, &attrib);
  if (wg->prop & NO_TITLE)
    {
      width = attrib.width - wg->bf_border * 2;
      height = attrib.height - wg->bf_border * 2;
    }
  else if (gl_epiwin.titlebar.position == TOP)
    {
      x = 0;
      y = wg->titlebar_height + wg->titlebar_border * 2
	- wg->bf_border;
      width = attrib.width - wg->bf_border * 2;
      height =  attrib.height - wg->titlebar_height
	- (wg->titlebar_border * 2) - wg->bf_border;
    }
  else if (gl_epiwin.titlebar.position == LEFT)
    {
      x = wg->titlebar_height + wg->titlebar_border * 2
	- wg->bf_border;
      y = 0;
      width = attrib.width - wg->bf_border - wg->titlebar_height
	- (wg->titlebar_border * 2);
      height = attrib.height - wg->bf_border * 2;
    }
  else if (gl_epiwin.titlebar.position == RIGHT)
    {
      x = 0;
      y = 0;
      width = attrib.width - wg->bf_border - wg->titlebar_height
	- (wg->titlebar_border * 2);
      height = attrib.height - wg->bf_border * 2;
    }
  else
    {
      x = 0;
      y = 0;
      width = attrib.width - wg->bf_border * 2;
      height =  attrib.height - wg->titlebar_height
	- (wg->titlebar_border * 2) - wg->bf_border;
      /*
	 width = attrib.width - wg->bf_border * 2;
	 height = attrib.height - wg->bf_border * 2;
	 */
    }
  tmp.bgcolor = BlackPixel(gl_dpy, gl_scr_info.screen);
  tmp.fgcolor = gl_epiwin.color_active.pixel;
  tmp.parent = parent;
  /*#warning Changer le cursor de border frame*/
  tmp.cursor = 0;
  tmp.border = wg->bf_border;
  tmp.x = x;
  tmp.y = y;
  tmp.width = width;
  tmp.height = height;
  w_tmp = add_btn(&tmp);
  attr.event_mask = ExposureMask | ButtonPressMask | ButtonReleaseMask
    | SubstructureRedirectMask | SubstructureNotifyMask | PointerMotionMask
      | KeyPressMask;
  XChangeWindowAttributes(gl_dpy, w_tmp, CWEventMask, &attr);
  return (w_tmp);
}

Window			add_titlebar_win(wg, parent)
t_win			*wg;
Window			parent;
{
  t_window_btn		tmp;
  XWindowAttributes	attrib;
  XSetWindowAttributes	attr;
  Window		w_tmp;
  int			width;
  int			height;
  int			x;
  int			y;
  
  if (wg->prop & NO_TITLE)
    return (0);
  x = 0;
  y = 0;
  width = 0;
  height = 0;
  XGetWindowAttributes(gl_dpy, parent, &attrib);
  if (gl_epiwin.titlebar.position == TOP)
    {
      x = 0;
      y = 0;
      width = attrib.width - 2 * wg->titlebar_border;
      height = wg->titlebar_height;
    }
  else if (gl_epiwin.titlebar.position == LEFT)
    {
      x = 0;
      y = 0;
      width = wg->titlebar_height;
      height = attrib.height - 2 * wg->titlebar_border;
    }
  else if (gl_epiwin.titlebar.position == RIGHT)
    {
      x = attrib.width - wg->titlebar_height - wg->titlebar_border * 2;
      y = 0;
      width = wg->titlebar_height;
      height = attrib.height - 2 * wg->titlebar_border;
    }
  else
    {
      x = 0;
      y = attrib.height - wg->titlebar_height
	- wg->titlebar_border * 2;
      width = attrib.width - 2 * wg->titlebar_border;
      height = wg->titlebar_height;
    }
  tmp.x = x;
  tmp.y = y;
  tmp.bgcolor = BlackPixel(gl_dpy, gl_scr_info.screen);
  tmp.fgcolor = gl_epiwin.titlebar.color_active.pixel;
  tmp.parent = parent;
  tmp.cursor = 0;
  tmp.border = wg->titlebar_border;
  tmp.width = width;
  tmp.height = height;
  w_tmp = add_btn(&tmp);
  attr.event_mask = ButtonPressMask | ButtonReleaseMask
    | ExposureMask | PointerMotionMask | KeyPressMask;
  XChangeWindowAttributes(gl_dpy, w_tmp, CWEventMask, &attr);
  return (w_tmp);
}


Window		add_btn(win_struct)
t_window_btn	*win_struct;
{
  Window	temp;
  
  temp = XCreateSimpleWindow(gl_dpy,
			     win_struct->parent,
			     win_struct->x,
			     win_struct->y,
			     win_struct->width,
			     win_struct->height,
			     win_struct->border,
			     win_struct->bgcolor,
			     win_struct->fgcolor);
  if (win_struct->cursor)
    XDefineCursor(gl_dpy, temp, win_struct->cursor);
  return (temp);
}

void	check_properties(wg_tmp)
t_win	*wg_tmp;
{  
  check_size_hints(wg_tmp);
  check_icon(wg_tmp);
}

void		check_icon(wg_tmp)
t_win		*wg_tmp;
{
  XWMHints	*wm_hints;
  
  if (gl_icon.showed)
    {
      wm_hints = XGetWMHints(gl_dpy, wg_tmp->win);
      if (wm_hints)
	{      
	  if (wm_hints->flags & IconWindowHint)
	    wg_tmp->iconify_win = wm_hints->icon_window;	  
	  else
	    wg_tmp->iconify_win = 0;
	  if (wm_hints->flags & IconMaskHint)
	    create_pixmap_mask(wg_tmp, &(wm_hints->icon_mask));
	  if (wm_hints->flags & IconPixmapHint)
	    create_icon_win_with_pixmap(wg_tmp, &(wm_hints->icon_pixmap));
	  if (wm_hints->flags & IconPositionHint);
	    /*	    printf("icon position\n");*/
	}
      else
	wg_tmp->iconify_win = 0;
    }
  else
    wg_tmp->iconify_win = 0;
}

void	create_icon_win_with_pixmap(wg_tmp, pixmap)
t_win	*wg_tmp;
Pixmap	*pixmap;
{
  int	width;
  int	height;
  
  if (pixmap != 0)
    {
      check_icon_size(pixmap, &width, &height);
      if (wg_tmp->iconify_win == 0)
	wg_tmp->iconify_win =
	  XCreateSimpleWindow(gl_dpy, gl_scr_info.root_win,
			      0, 0, width, height, 0,
			      BlackPixel(gl_dpy,
					 gl_scr_info.screen),
			      WhitePixel(gl_dpy,
					 gl_scr_info.screen));
      else
	XResizeWindow(gl_dpy, wg_tmp->iconify_win, width, height);      
      XSetWindowBackgroundPixmap(gl_dpy, wg_tmp->iconify_win, *pixmap);
      iconify_win_set_attrib(wg_tmp);
    }  
}

void	create_pixmap_mask(wg_tmp, shape_pixmap)
t_win	*wg_tmp;
Pixmap	*shape_pixmap;
{
  XShapeCombineMask(gl_dpy, wg_tmp->iconify_win, ShapeBounding,
		    0, 0, *shape_pixmap, ShapeSet);
}

void			iconify_win_set_attrib(wg_tmp)
t_win			*wg_tmp;
{
  XSetWindowAttributes	attr;
  
  attr.event_mask = ButtonPressMask | ButtonReleaseMask
    | EnterWindowMask | LeaveWindowMask | ExposureMask | PointerMotionMask;
  XChangeWindowAttributes(gl_dpy, wg_tmp->iconify_win, CWEventMask, &attr);
}

void			check_icon_size(pixmap, width, height)
Pixmap			*pixmap;
int			*width;
int			*height;
{
  Window		dummy;
  int			dummy_int;
  
  XGetGeometry(gl_dpy, *pixmap, &dummy, &dummy_int, &dummy_int, width, 
	       height, &dummy_int, &dummy_int);
}

void		check_size_hints(wg_tmp)
t_win		*wg_tmp;
{
  XSizeHints	size_hints;
  long		toto;
  
  XGetWMNormalHints(gl_dpy, wg_tmp->win, &size_hints, &toto);
  if (size_hints.flags & PResizeInc)
    {
      if (size_hints.width_inc == 0)
	wg_tmp->hints.width_inc = 1;
      else
	wg_tmp->hints.width_inc = size_hints.width_inc;
      if (size_hints.height_inc == 0)
	wg_tmp->hints.height_inc = 1;
      else
	wg_tmp->hints.height_inc = size_hints.height_inc;
    }
  else
    {
      wg_tmp->hints.width_inc = 1;
      wg_tmp->hints.height_inc = 1;
    }
  if (size_hints.flags & PMinSize)
    {
      if (size_hints.min_width  == 0)
	wg_tmp->hints.min_width = 1;
      else
	wg_tmp->hints.min_width = size_hints.min_width;
      if (size_hints.min_height  == 0)
	wg_tmp->hints.min_height = 1;
      else
	wg_tmp->hints.min_height = size_hints.min_height;
    }
  else
    {
      wg_tmp->hints.min_width = 1;
      wg_tmp->hints.min_height = 1;
    }
  if (size_hints.flags & PMaxSize)
    {
      if (size_hints.max_width  == 0)
	wg_tmp->hints.max_width = 1;
      else
	wg_tmp->hints.max_width = size_hints.max_width;
      if (size_hints.max_height == 0)
	wg_tmp->hints.max_height = 1;
      else
	wg_tmp->hints.max_height = size_hints.max_height;
    }
  else
    {
      wg_tmp->hints.max_width = -1;
      wg_tmp->hints.max_height = -1;
    }
  verify_max_min(wg_tmp);
}

void	verify_max_min(wg_tmp)
t_win	*wg_tmp;
{
  int	tmp;
  
  if (wg_tmp->hints.max_width != -1)
    if (wg_tmp->hints.max_width < wg_tmp->hints.min_width)
      {
	tmp = wg_tmp->hints.max_width;
	wg_tmp->hints.max_width = wg_tmp->hints.min_width;
	wg_tmp->hints.min_width = tmp;
      }
  if (wg_tmp->hints.max_height != -1)
    if (wg_tmp->hints.max_height < wg_tmp->hints.min_height)
      {
	tmp = wg_tmp->hints.max_height;
	wg_tmp->hints.max_height = wg_tmp->hints.min_height;
	wg_tmp->hints.min_height = tmp;
      }
}



